from functools import wraps

from .core import Locator
from .injector import Injector


def func_inject(
        dependencies: Locator = None,
        injector: Injector = None,
):
    if injector:
        assert not dependencies,\
            '"{}" decorator requires one argument only, two provided.'.format(func_inject.__name__)
        inj = injector
    elif dependencies:
        inj = Injector(dependencies)
    else:
        raise AssertionError('"{}" decorator requires one single argument, none provided.'.format(func_inject.__name__))

    def get_func_wrapper(func):

        def func_wrapper(*args, **kwargs):
            return inj.call(func, *args, **kwargs)
        return func_wrapper

    return get_func_wrapper


def inject(init_method):

    assert init_method.__name__ == '__init__', \
        '"{}" decorator must precede __init__ method.'.format(inject.__name__)

    @wraps(init_method)
    def init_wrapper(self, *args, dependencies=None, **kwargs):
        # Inject
        if dependencies:
            for name, value in dependencies.items():
                setattr(self, name, value)

        # run __init__
        init_method(self, *args, **kwargs)

    return init_wrapper



import logging
from io import StringIO

from unittest import TestCase

from wirinj import Definitions
from wirinj.decorators import func_inject, inject
from wirinj.definition import Autowiring
from tests.example_classes import Mike, Engine
from tests.example_definitions import world_one_deps


class DecoratorsTest(TestCase):
    def test_func_inject(self):
        engine = Engine(dependencies={'mount_sound': 'Brrrooommm'})
        self.assertEqual(engine.mount_sound, 'Brrrooommm')

    def test_inject(self):
        log = StringIO()
        logging.basicConfig(stream=log, level=logging.INFO, format='%(message)s')

        @func_inject(world_one_deps)
        def do(mike: Mike):
            mike.deliver(100, 5, False)

        do()

        log_str = log.getvalue()
        self.assertMultiLineEqual(
            log_str,
            'Picking pets up:  MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW MEOW\nVan is built:  RRRRoarrr plaf plaf plaf plaf plaf plaf plaf plaf pffff pffff pffff pffff\nUploading to the van: Cat Cat Cat Cat Cat Cat Cat Cat Cat Cat\nVan is built:  RRRRoarrr plaf plaf plaf plaf plaf plaf plaf plaf pffff pffff pffff pffff\nUploading to the van: Cat Cat Cat Cat Cat Cat Cat Cat Cat Cat\nVan is built:  RRRRoarrr plaf plaf plaf plaf plaf plaf plaf plaf pffff pffff pffff pffff\nUploading to the van: Cat Cat Cat Cat Cat Cat Cat Cat Cat Cat\nVan is built:  RRRRoarrr plaf plaf plaf plaf plaf plaf plaf plaf pffff pffff pffff pffff\nUploading to the van: Cat Cat Cat Cat Cat Cat Cat Cat Cat Cat\nVan is built:  RRRRoarrr plaf plaf plaf plaf plaf plaf plaf plaf pffff pffff pffff pffff\nUploading to the van: Cat Cat Cat Cat Cat Cat Cat Cat Cat Cat\nVan is built:  RRRRoarrr plaf plaf plaf plaf plaf plaf plaf plaf pffff pffff pffff pffff\nUploading to the van: Cat Cat Cat Cat Cat Cat Cat Cat Cat Cat\nVan is built:  RRRRoarrr plaf plaf plaf plaf plaf plaf plaf plaf pffff pffff pffff pffff\nUploading to the van: Cat Cat Cat Cat Cat Cat Cat Cat Cat Cat\nVan is built:  RRRRoarrr plaf plaf plaf plaf plaf plaf plaf plaf pffff pffff pffff pffff\nUploading to the van: Cat Cat Cat Cat Cat Cat Cat Cat Cat Cat\nVan is built:  RRRRoarrr plaf plaf plaf plaf plaf plaf plaf plaf pffff pffff pffff pffff\nUploading to the van: Cat Cat Cat Cat Cat Cat Cat Cat Cat Cat\nVan is built:  RRRRoarrr plaf plaf plaf plaf plaf plaf plaf plaf pffff pffff pffff pffff\nUploading to the van: Cat Cat Cat Cat Cat Cat Cat Cat Cat Cat\nVan goes 5 miles\nVan goes 5 miles\nVan goes 5 miles\nVan goes 5 miles\nVan goes 5 miles\nVan goes 5 miles\nVan goes 5 miles\nVan goes 5 miles\nVan goes 5 miles\nVan goes 5 miles\n10 pets delivered\n10 pets delivered\n10 pets delivered\n10 pets delivered\n10 pets delivered\n10 pets delivered\n10 pets delivered\n10 pets delivered\n10 pets delivered\n10 pets delivered\nVan commes back\nVan commes back\nVan commes back\nVan commes back\nVan commes back\nVan commes back\nVan commes back\nVan commes back\nVan commes back\nVan commes back\n',
            'Mismatch in output of test_classes',
        )

    def test_double_inject_on_subclass(self):
        class Qux:
            pass

        class Foo:
            pass

        class Bar:

            def __deps__(self, foo: Foo, foo2: Foo, **_):
                pass

            @inject
            def __init__(self):
                pass

        class Baz(Bar):

            def __deps__(self, qux: Qux, **_):
                pass

            @inject
            def __init__(self):
                super().__init__()

        @func_inject(Definitions(Autowiring()))
        def do(baz: Baz):
            self.assertIsNotNone(baz.foo)
            self.assertIsNotNone(baz.foo2)
            self.assertIsNotNone(baz.qux)
            self.assertIs(baz.foo, baz.foo2)
        do()



from .injector import Injector
__all__ = [Injector.__name__]

from .core import Arg, Dependency, Locator
__all__.extend([Arg.__class__, Dependency.__class__, Locator.__class__])

from .decorators import inject, func_inject, Injector
__all__.extend([inject.__name__, func_inject.__name__, Injector.__name__])

from .definition import Definitions, DependencyBuilder, Singleton, Factory, Instance, CustomSingleton, CustomInstance, CustomFactory, \
    Autowiring, AutowiringReport
__all__.extend([Definitions.__name__, DependencyBuilder.__name__, Singleton.__name__, Factory.__name__,
                Instance.__name__, CustomSingleton.__name__, CustomInstance.__name__, CustomFactory.__name__,
                Autowiring.__name__, AutowiringReport.__name__])

from .locator import Locator, LocatorCache, LocatorChain
__all__.extend([Locator.__name__, LocatorCache.__name__, LocatorChain.__name__])



from logging import ERROR, INFO, DEBUG
from typing import Union, Sequence, List, Callable, Optional, Dict, Tuple

from .core import logger, Arg, Dependency, NotSet, Locator, SEPARATOR_OPEN, SEPARATOR_CLOSE, InstanceArgs, \
    filter_explicit_args
from .errors import MissingDependenciesError
from .introspection import get_func_args
from .sys import get_subclassing_factory, get_func_factory


class InjectType(type):
    def __str__(self):
        return 'Inject'


class Inject(metaclass=InjectType):
    pass


class NotFoundType(type):
    def __str__(self):
        return 'NotFound'

class NotFound(metaclass=NotFoundType):
    pass


class FailedType(type):
    def __str__(self):
        return 'Failed'

class Failed(metaclass=FailedType):
    pass


INJECT = '__inject__'


class CreationNode:
    def __init__(self,
                 arg: Arg, dep: Union[Dependency, NotFoundType, None],
                 childs: Sequence['CreationNode'] = (),
                 ):
        self.arg = arg
        self.dep = dep
        self.childs = childs
        self.instance = None

    def get_params(self):
        result = {}
        for child in self.childs:
            result[child.arg.name] = child.instance
        return result

    def __str__(self) -> str:
        return "{}{}".format(
            str(self.arg),
            ' *** {} ***'.format(str(self.dep)) if self.dep is NotFound else '',
        )


class CreationNodeReference:
    def __init__(self, creation_path: Sequence[CreationNode], node: CreationNode):
        self.creation_path = creation_path
        self.node = node

    def __str__(self) -> str:

        path = ''
        for entry in self.creation_path:
            path += '    '

        return "{}{}".format(
            path,
            str(self.node),
        )


def get_tree_index(root: CreationNode, creation_path=()) -> Sequence[CreationNodeReference]:
    result = []
    for child in root.childs:
        result += get_tree_index(child, (list(creation_path) + [root.arg]))

    result += [CreationNodeReference(creation_path, root)]

    return result


def get_creation_args(node_childs: Sequence[CreationNode]) -> Dict:
    result = {}
    for child in node_childs:
        result[child.arg.name] = child.instance
    return result



def log_creation_tree(tree: CreationNode, log_level=INFO, msg = None):

    if not logger.isEnabledFor(log_level):
        return

    dep_flat = get_tree_index(tree)

    logger.log(log_level, SEPARATOR_OPEN)
    if msg:
        logger.log(log_level, msg)
    for node_ref in dep_flat:
        logger.log(log_level, str(node_ref))
    logger.log(log_level, SEPARATOR_CLOSE)


def _after_tree_creation(success: bool, tree: CreationNode):
    if success:
        log_creation_tree(tree, DEBUG)
    else:
        log_creation_tree(tree, ERROR, 'Missing dependencies:')
        raise MissingDependenciesError('Missing dependencies.')


def creation_path_as_text(parent_path: Sequence[Arg], creation_node: CreationNode, suffix=''):
    path = ''
    for entry in parent_path[0:-1]:
        path += str(entry) + ' -> '

    path += str(creation_node.arg)

    if suffix:
        path += ' [{}]'.format(suffix)

    return path



def has_a_valid_default(arg: Arg):
    if arg.default is NotSet:
        return False

    if arg.cls != NotSet and arg.default in [INJECT, Inject]:
        return False

    return True


class DefaultDependency(Dependency):

    def __init__(self, default):
        self.default = default

    def get_instance(self, instance_args=None, **deps):
        return self.default


class Injector:

    def __init__(self, dependencies: Locator, subclassing_factory=True):
        self.subclassing_factory = subclassing_factory
        dependencies.initialize(self)
        self.locator = dependencies

    def get(self, cls, *args, **kwargs):
        success, root = self._create_node((), Arg(None, cls), InstanceArgs(args, kwargs))
        _after_tree_creation(success, root)
        return root.instance

    def call(self, func: Callable, *args, **kwargs):
        injected_args = self._get_function_args(func, args, kwargs)
        return func(*args, **{**injected_args, **kwargs})

    def get_wrapper(self, func: Callable):
        def func_wrapper(*args, **kwargs):
            injected_args = self._get_function_args(func, args, kwargs)
            return func(*args, **{**injected_args, **kwargs})
        return func_wrapper

    def get_factory(self, cls):
        if self.subclassing_factory:
            return get_subclassing_factory(cls, self.get)
        else:
            return get_func_factory(cls, self.get)

    def _get_function_args(self, func: Callable, args, kwargs):
        fn_args = get_func_args(func)
        injectable_args = filter_explicit_args(fn_args, args, kwargs)
        return self._create_virtual_node(injectable_args, Arg(func.__name__, func.__class__, NotSet))

    def _create_childs(self, arg_list: Optional[Sequence[Arg]], creation_path: Sequence[Arg]) -> (bool, List[CreationNode]):
        """
        @return: Returns two values. The first value is True if the instantiation of all childs succeded. The second
        is the CreationNode list.
        """
        if not arg_list:
            return True, ()

        childs = []
        success = True
        for arg in arg_list:
            child_success, child = self._create_node(creation_path, arg)
            if not child_success:
                success = False
            childs.append(child)
        return success, childs

    def _create_node(self, parent_path: Sequence[Arg], arg: Arg, instance_args: Optional[InstanceArgs] = None) -> Tuple[bool, CreationNode]:
        """
        @return: Returns two values. The first value is True if the instantiation succeded. The second is the new
        node. When the Locator cannot find a dependency, the field 'dep' will be set as NotFound.
        """
        current_path = tuple(parent_path) + (arg,)

        # Find in the locator
        dep = self.locator.get(current_path)
        if not dep:
            if has_a_valid_default(arg):
                dep = DefaultDependency(arg.default)
            else:
                return False, CreationNode(current_path[-1], NotFound)

        # Get dependency args
        dep_args = dep.get_dependencies()

        # With args
        if dep_args:
            # Update path with actual class
            cls = dep.get_class()
            if cls is not NotSet and cls != arg.cls:
                current_path = list(parent_path) + [Arg(arg.name, cls, arg.default)]

            # Create childs
            childs_success, childs = self._create_childs(dep_args, current_path)

            # Creation params
            params = get_creation_args(childs)

        # With no args
        else:
            childs =()
            params = {}
            childs_success = True

        # New CreationNode
        creation_node = CreationNode(current_path[-1], dep, childs)

        # Missing childs dependencies. Return Node with no instance.
        if not childs_success:
            return False, creation_node

        # All dependencies fulfilled. Create instance
        try:
            creation_node.instance = dep.get_instance(instance_args, **params)
        except BaseException as ex:
            creation_node.instance = Failed
            logger.fatal(SEPARATOR_OPEN)
            logger.fatal('Instantiation ERROR:')
            logger.fatal(creation_path_as_text(parent_path, creation_node, ex.__class__.__name__))
            logger.fatal(SEPARATOR_CLOSE)
            raise ex

        # Return full node
        return True, creation_node


    def _create_virtual_node(self, args: Sequence[Arg], root_arg: Arg):

        # Create tree of dependencies
        success, childs = self._create_childs(args, [root_arg])
        root = CreationNode(root_arg, None, childs)
        _after_tree_creation(success, root)
        return root.get_params()




from unittest import TestCase

from wirinj import Definitions, Injector
from wirinj.decorators import func_inject
from wirinj.definition import CustomSingleton, Singleton, CustomInstance


class TestCustomSingletonDependency(TestCase):

    def test_injection(self):
        class Foo:
            def __init__(self, bar):
                self.bar = bar

        def foo_creator(bar) -> Foo:
            return Foo(bar)

        deps = Definitions({
            'bar': 'Hola',
            Foo: CustomSingleton(foo_creator),
        })

        @func_inject(deps)
        def do(foo: Foo):
            self.assertIsNotNone(foo)
            self.assertIsNotNone(foo.bar)

        do()

class TestCustomInstanceDependency(TestCase):
    def test_get_with_params(self):
        class Baz:
            pass

        class Foo:
            def __init__(self, bar, baz):
                self.bar = bar
                self.baz = baz

        def foo_creator(baz: Baz):
            def creator(bar):
                return Foo(bar, baz)

            return creator

        injector = Injector(Definitions({
            Baz: Singleton(),
            Foo: CustomInstance(foo_creator, True),
        }))

        foo = injector.get(Foo, 'Ping')
        self.assertIs(foo.bar, 'Ping')
        self.assertIsNotNone(foo.baz)


        foo2 = injector.get(Foo, bar='Pong')
        self.assertIs(foo2.bar, 'Pong')
        self.assertIsNotNone(foo2.baz)

        pass



from inspect import getfullargspec, FullArgSpec
from typing import Type, Sequence, Callable, Optional

from .core import Arg, NotSet, InjectionType, DEPS_METHOD, InstanceArgs, NotSetType


def is_builtin_cls(annotation) -> [NotSetType, bool]:
    return NotSet if annotation is NotSet else annotation.__module__ == 'builtins'


def get_deps_from_argspec(spec: FullArgSpec) -> Sequence[Arg]:
    result = []

    defaults_idx_base = (0 if spec.defaults is None else len(spec.defaults)) - len(spec.args)
    n = 0
    for i, name in enumerate(spec.args):
        if name == 'self':
            continue
        default_idx = defaults_idx_base + i
        default = spec.defaults[default_idx] if default_idx >= 0 else NotSet
        annotation = spec.annotations.get(name, NotSet)
        result.append(Arg(name, annotation, default))
        n += 1
    return result



def get_func_result(func: Callable) -> Sequence[Arg]:
    assert isinstance(func, Callable), \
        '"{}" must be Callable'.format(func.__name__)
    spec = getfullargspec(func)
    return spec.annotations.get('return', NotSet)


def get_func_args(func: Callable) -> Sequence[Arg]:

    assert isinstance(func, Callable), \
        '"{}" must be Callable'.format(func.__name__)

    spec = getfullargspec(func)
    return get_deps_from_argspec(spec)


def get_method_args(cls, method_name) -> Sequence[Arg]:

    method = getattr(cls, method_name)

    assert method, 'Missing method {1} on class {0}'.format(cls.__name__, method_name)

    assert isinstance(method, Callable), \
        '"{0}.{1}" must be Callable'.format(cls.__name__, method_name)

    spec = getfullargspec(method)
    return get_deps_from_argspec(spec)



def get_class_injection_type(cls) -> InjectionType:
    init_method = getattr(cls, '__init__')

    if not init_method:
        return InjectionType.none

    spec = getfullargspec(init_method)

    try:
        if spec.kwonlyargs and spec.kwonlyargs.index('dependencies') >= 0:
            return InjectionType.deps
    except ValueError:
        pass

    return InjectionType.init


def get_class_deps(cls):
    args = []

    # Base deps
    for base_cls in cls.__bases__:
        if base_cls.__module__ != 'builtins':
            args += get_class_deps(base_cls)

    # Current deps
    method = getattr(cls, DEPS_METHOD, None)
    if method:
        args += get_method_args(cls, DEPS_METHOD)

    return args


def get_class_dependencies(cls) -> Sequence[Arg]:

    injection_type = get_class_injection_type(cls)

    if injection_type == InjectionType.none:
        return ()
    elif injection_type == InjectionType.init:
        return get_method_args(cls, '__init__')
    elif injection_type == InjectionType.deps:
        return get_class_deps(cls)
    else:
        raise ValueError()


def instantiate_class(cls, instance_args: Optional[InstanceArgs] = None, **deps):

    type = get_class_injection_type(cls)

    if instance_args:
        args = instance_args.args
        kwargs = instance_args.kwargs
    else:
        args = ()
        kwargs = {}

    if type == InjectionType.none:
        return cls()
    elif type == InjectionType.init:
        return cls(*args, **{**deps, **kwargs})
    elif type == InjectionType.deps:
        return cls(*args, dependencies=deps, **kwargs)
    else:
        raise ValueError()



from abc import abstractmethod, ABCMeta
from inspect import isabstract
from typing import Sequence, List, Callable, Optional, Dict, Union

from .core import Arg, NotSet, Locator, Dependency, SEPARATOR_OPEN, SEPARATOR_CLOSE
from .dependency import FactoryDependency, InstanceDependency, SingletonWrapper, ValueDependency, \
    CustomInstanceDependency
from .injector import Injector
from .locator import LocatorCache, LocatorChain
from .sys import is_typing_type, get_typing_args


class DefinitionFinder:
    def __init__(self, definitions: Dict):
        assert isinstance(definitions, dict)
        self.definitions = definitions

    def __call__(self, creation_path: List[Arg]):

        result_key = None
        result_value = None
        def_len = 0

        for key, value in self.definitions.items():
            definition = key if isinstance(key, list) or isinstance(key, tuple) else (key,)

            if len(definition) > len(creation_path):
                continue

            if len(definition) <= def_len:
                continue

            match = True
            for i, def_item in enumerate(reversed(definition)):
                path_item = creation_path[-i-1]
                if not(isinstance(def_item, str) and def_item == path_item.name or def_item == path_item.cls):
                    match = False
                    break
            if match:
                result_key = key
                result_value = value
                def_len = len(definition)

        return (NotSet, NotSet) if def_len == 0 else (result_key, result_value)


class DependencyBuilder:

    @abstractmethod
    def create(self, creation_path: List[Arg], injector: Injector):
        pass


class CustomInstance(DependencyBuilder):
    def __init__(self, creator: Callable, wrapper_func: bool = False, cls=None):
        self.creator = creator
        self.wrapper_func = wrapper_func
        self.cls = cls

    def create(self, creation_path: List[Arg], injector: Injector):
        return CustomInstanceDependency(self.creator, self.wrapper_func, self.cls)


class CustomSingleton(DependencyBuilder):
    def __init__(self, creator: Callable, wrapper_func: bool = False, cls=None):
        self.creator = creator
        self.wrapper_func = wrapper_func
        self.cls = cls

    def create(self, creation_path: List[Arg], injector: Injector):
        return SingletonWrapper(CustomInstanceDependency(self.creator, self.wrapper_func, self.cls))

CustomFactory = CustomSingleton


class Factory(DependencyBuilder):
    def __init__(self, cls=None):
        self.cls = cls

    def create(self, creation_path: List[Arg], injector: Injector):
        if self.cls is None:
            last = creation_path[-1]
            assert is_typing_type(last.cls),\
                'Factory without params needs Type[YourClass] as last element in the definition path'
            cls = get_typing_args(last.cls)
        else:
            cls = self.cls

        return SingletonWrapper(FactoryDependency(cls, injector))


class Instance(DependencyBuilder):
    def __init__(self, cls=None):
        self.cls = cls

    def create(self, creation_path: List[Arg], injector: Injector):
        if self.cls is None:
            last = creation_path[-1]
            assert isinstance(last.cls, type),\
                'Instance without params needs YourClass as last element in the definition path'
            cls = last.cls
        else:
            cls = self.cls

        return InstanceDependency(cls)


class Singleton(DependencyBuilder):
    def __init__(self, cls=None):
        self.cls = cls

    def create(self, creation_path: List[Arg], injector: Injector):
        if self.cls is None:
            last = creation_path[-1]
            assert isinstance(last.cls, type), \
                'Singleton without params needs YourClass as last element in the definition path'
            cls = last.cls
        else:
            cls = self.cls

        return SingletonWrapper(InstanceDependency(cls))



class DefinitionLocator(Locator):
    def __init__(self, definitions: Dict):
        self.finder = DefinitionFinder(definitions)
        self.injector = None  # type: Optional[Injector]

    def initialize(self, injector):
        self.injector = injector

    def get(self, creation_path: List[Arg]) -> Optional[Dependency]:
        assert self.injector is not None

        _ , value = self.finder(creation_path)
        if value is NotSet:
            return None
        if isinstance(value, DependencyBuilder):
            return value.create(creation_path, self.injector)
        else:
            return ValueDependency(value)


class AutowiringReportBase(metaclass=ABCMeta):

    @abstractmethod
    def add(self, line):
        pass

    @abstractmethod
    def get(self):
        pass

class NullAutowiringReport(AutowiringReportBase):

    def add(self, line):
        pass

    def get(self):
        return ''


class AutowiringReport(AutowiringReportBase):

    def __init__(self):
        self.lines = []

    def add(self, line):
        self.lines.append(line)

    def get(self):
        if not self.lines:
            return ''

        lines = ''
        for line in self.lines:
            lines += '    {},\n'.format(line)
        return '{0}\n{1}\n\n{2}({{\n{3}}}),\n{4}'.format(
            SEPARATOR_OPEN,
            'Autowiring report:',
            DefinitionLocator.__name__,
            lines,
            SEPARATOR_CLOSE,
        )


class Autowiring(Locator):

    def __init__(self, report: Optional[AutowiringReport] = None):
        self.report = report if report else NullAutowiringReport()
        self.singletons = {}

    def initialize(self, injector):
        self.injector = injector

    def get(self, creation_path: Sequence[Arg]) -> Optional[Dependency]:
        arg = creation_path[-1]

        if arg.cls is NotSet:
            return None

        try:
            return self.singletons[arg.cls]
        except KeyError:
            pass

        # Something annotated with Type[] is a factory
        if is_typing_type(arg.cls):
            type_cls = get_typing_args(arg.cls)
            self.report.add('Type[{}]: Factory()'.format(type_cls.__name__))
            factory_dep = SingletonWrapper(FactoryDependency(type_cls, self.injector))
            self.singletons[arg.cls] = factory_dep
            return factory_dep

        # If a non abstract class
        elif not isabstract(arg.cls):

            # With a name, this comes probably from a function arg and therefore it is presumably a singleton
            if arg.name:
                self.report.add('{}: Singleton()'.format(arg.cls.__name__))
                singleton_dep = SingletonWrapper(InstanceDependency(arg.cls))
                self.singletons[arg.cls] = singleton_dep
                return singleton_dep

            # Without a name, this comes probably from a factory call and therefore it is presumably an instance
            else:
                self.report.add('{}: Instance()'.format(arg.cls.__name__))
                return InstanceDependency(arg.cls)
        return None


class Definitions(Locator):
    def __init__(self, *args: Union[Locator, dict], cached=True):

        assert len(args), 'You have to provide at least one dependency locator'

        locators = []
        for arg in args:
            if isinstance(arg, Locator):
                locators.append(arg)

            elif isinstance(arg, dict):
                locators.append(DefinitionLocator(arg))

        uncached = LocatorChain(locators) if len(locators) > 1 else locators[0]
        self.locator = LocatorCache(uncached) if cached else uncached

    def initialize(self, injector):
        self.locator.initialize(injector)

    def get(self, creation_path: Sequence[Arg]) -> Optional[Dependency]:
        return self.locator.get(creation_path)



from abc import abstractmethod, ABCMeta
from enum import Enum
from logging import getLogger
from typing import Optional, Union, Any, Sequence

import wirinj
from .sys import get_cls_name

logger = getLogger(wirinj.__name__)


SEPARATOR_OPEN  = '--------------- wirinj ---------------'
SEPARATOR_CLOSE = '--------------------------------------'

DEPS_METHOD = '__deps__'


USE_SUBCLASSING_FACTORY = True

class NotSetType(type):
    def __str__(self):
        return 'NotSet'


class NotSet(metaclass=NotSetType):
    pass


class InjectionType(Enum):
    init = 1
    deps = 2
    none = 3


class InstanceArgs:
    def __init__(self, args, kwargs):
        self.args = args
        self.kwargs = kwargs


class Arg:
    def __init__(self, name: Optional[str], cls: Union[NotSetType, Any] = NotSet, default: Union[NotSetType, Any] = NotSet):
        self.name = name
        self.cls = cls
        self.default = default

    def __str__(self) -> str:
        return "{}{}{}".format(
            str('' if self.name is None else self.name),
            '' if self.cls is NotSet else ':{}'.format(get_cls_name(self.cls)),
            '' if self.default is NotSet else '={}'.format(str(self.default)),
        )

    def __eq__(self, other):
        return hash(self) == hash(other)

    def __hash__(self):
        return hash((hash(self.name), hash(self.cls), hash(self.default)))


class Dependency(metaclass=ABCMeta):

    def get_class(self) -> Union[Any, NotSet]:
        return NotSet

    def get_dependencies(self) -> Optional[Sequence[Arg]]:
        return None

    @abstractmethod
    def get_instance(self, instance_args: InstanceArgs = None, **deps):
        pass


class Locator(metaclass=ABCMeta):

    def initialize(self, injector):
        pass

    @abstractmethod
    def get(self, creation_path: Sequence[Arg]) -> Optional[Dependency]:
        pass


def filter_explicit_args(arg_list: Sequence[Arg], args, kwargs):
    result = []
    for i, arg in enumerate(arg_list):
        if i < len(args):
            continue

        if arg.name in kwargs and arg.default != kwargs[arg.name]:
            continue

        result.append(arg)

    return result


def join_arg_sequences(*arg_sequences: Sequence[Arg]):
    result = []
    index = {}
    for arg_sequence in reversed(arg_sequences):
        for arg in arg_sequence:
            if index.get(arg.name):
                continue
            index[arg.name] = True
            result += arg
    return result
